@@ -5,13 +5,15 @@ use std::{
 
 use crate::{
     counter::Counter,
-    error::{OperationalError, PolarResult, RuntimeError},
+    error::{invalid_state_error, RuntimeError},
     events::ResultEvent,
+    filter::singleton,
     terms::*,
 };
 
 use serde::{Deserialize, Serialize};
 
+type Result<A> = core::result::Result<A, RuntimeError>;
 type Id = u64;
 type VarId = Id;
 type TypeName = String;
@@ -127,7 +129,7 @@ pub fn build_filter_plan(
     partial_results: PartialResults,
     variable: &str,
     class_tag: &str,
-) -> PolarResult<FilterPlan> {
+) -> Result<FilterPlan> {
     FilterPlan::build(types, partial_results, variable, class_tag)
 }
 
@@ -142,7 +144,7 @@ impl From<Term> for Constraint {
 }
 
 impl VarInfo {
-    fn from_op(op: &Operation) -> PolarResult<Self> {
+    fn from_op(op: &Operation) -> Result<Self> {
         Self::default().process_exp(op)
     }
 
@@ -210,29 +212,32 @@ impl VarInfo {
         }
     }
 
-    fn do_and(self, args: &[Term]) -> PolarResult<Self> {
+    fn do_and(self, args: &[Term]) -> Result<Self> {
         args.iter().fold(Ok(self), |this, arg| {
             this?.process_exp(arg.value().as_expression().unwrap())
         })
     }
 
-    fn do_dot(mut self, lhs: &Term, rhs: &Term) -> PolarResult<Self> {
+    fn do_dot(mut self, lhs: &Term, rhs: &Term) -> Result<Self> {
         self.dot_var(lhs, rhs);
         Ok(self)
     }
 
-    fn do_isa(mut self, lhs: &Term, rhs: &Term) -> PolarResult<Self> {
+    fn do_isa(mut self, lhs: &Term, rhs: &Term) -> Result<Self> {
         match rhs.value().as_pattern() {
             Ok(Pattern::Instance(i)) if i.fields.fields.is_empty() => {
                 let lhs = self.symbolize(lhs);
                 self.types.push((lhs, i.tag.0.clone()));
                 Ok(self)
             }
-            _ => err_unimplemented(format!("Unsupported specializer: {}", rhs.to_polar())),
+            _ => unsupported_op_error(Operation {
+                operator: Operator::Isa,
+                args: vec![lhs.clone(), rhs.clone()],
+            }),
         }
     }
 
-    fn do_unify(mut self, left: &Term, right: &Term) -> PolarResult<Self> {
+    fn do_unify(mut self, left: &Term, right: &Term) -> Result<Self> {
         match (self.undot(left), self.undot(right)) {
             (Value::Variable(l), Value::Variable(r)) => {
                 self.cycles.push((l, r));
@@ -246,15 +251,14 @@ impl VarInfo {
             // 1 = 1 is irrelevant for data filtering, other stuff seems like an error.
             // @NOTE(steve): Going with the same not yet supported message but if this is
             // coming through it's probably a bug in the simplifier.
-            _ => err_unimplemented(format!(
-                "Unsupported unification: {} = {}",
-                left.to_polar(),
-                right.to_polar()
-            )),
+            _ => unsupported_op_error(Operation {
+                operator: Operator::Unify,
+                args: vec![left.clone(), right.clone()],
+            }),
         }
     }
 
-    fn do_neq(mut self, left: &Term, right: &Term) -> PolarResult<Self> {
+    fn do_neq(mut self, left: &Term, right: &Term) -> Result<Self> {
         match (self.undot(left), self.undot(right)) {
             (Value::Variable(l), Value::Variable(r)) => {
                 self.uncycles.push((l, r));
@@ -270,15 +274,14 @@ impl VarInfo {
                 self.uncycles.push((l, r));
                 Ok(self)
             }
-            _ => err_unimplemented(format!(
-                "Unsupported comparison: {} != {}",
-                left.to_polar(),
-                right.to_polar()
-            )),
+            _ => unsupported_op_error(Operation {
+                operator: Operator::Neq,
+                args: vec![left.clone(), right.clone()],
+            }),
         }
     }
 
-    fn do_in(mut self, left: &Term, right: &Term) -> PolarResult<Self> {
+    fn do_in(mut self, left: &Term, right: &Term) -> Result<Self> {
         match (self.undot(left), self.undot(right)) {
             (Value::Variable(l), Value::Variable(r)) => {
                 self.in_relationships.push((l, r));
@@ -288,16 +291,15 @@ impl VarInfo {
                 self.contained_values.push((Term::from(val), var));
                 Ok(self)
             }
-            _ => err_unimplemented(format!(
-                "Unsupported `in` check: {} in {}",
-                left.to_polar(),
-                right.to_polar()
-            )),
+            _ => unsupported_op_error(Operation {
+                operator: Operator::In,
+                args: vec![left.clone(), right.clone()],
+            }),
         }
     }
 
     /// Process an expression in the context of this VarInfo. Just does side effects.
-    fn process_exp(self, exp: &Operation) -> PolarResult<Self> {
+    fn process_exp(self, exp: &Operation) -> Result<Self> {
         use Operator::*;
         let args = &exp.args;
         match exp.operator {
@@ -307,29 +309,20 @@ impl VarInfo {
             Neq if args.len() == 2 => self.do_neq(&args[0], &args[1]),
             In if args.len() == 2 => self.do_in(&args[0], &args[1]),
             Unify | Eq | Assign if args.len() == 2 => self.do_unify(&args[0], &args[1]),
-            _ => err_unimplemented(format!(
-                "the expression {:?}/{} is not supported for data filtering",
-                exp.operator,
-                exp.args.len()
-            )),
+            _ => unsupported_op_error(exp.clone()),
         }
     }
 }
 
-fn unregistered_field_error<A>(var_type: &str, field: &str) -> PolarResult<A> {
+pub fn unsupported_op_error<A>(operation: Operation) -> Result<A> {
+    Err(RuntimeError::DataFilteringUnsupportedOp { operation })
+}
+
+pub fn unregistered_field_error<A>(var_type: &str, field: &str) -> Result<A> {
     Err(RuntimeError::DataFilteringFieldMissing {
         var_type: var_type.to_string(),
         field: field.to_string(),
-    }
-    .into())
-}
-
-fn err_invalid<A>(msg: String) -> PolarResult<A> {
-    Err(OperationalError::InvalidState { msg }.into())
-}
-
-fn err_unimplemented<A>(msg: String) -> PolarResult<A> {
-    Err(OperationalError::Unimplemented { msg }.into())
+    })
 }
 
 impl FilterPlan {
@@ -338,7 +331,7 @@ impl FilterPlan {
         partial_results: PartialResults,
         var: &str,
         class_tag: &str,
-    ) -> PolarResult<FilterPlan> {
+    ) -> Result<FilterPlan> {
         // @NOTE(steve): Just reading an env var here sucks (see all the stuff we had to do
         // to get POLAR_LOG to work in all libs, wasm etc...) but that's what I'm doing today.
         // At some point surface this info better.
@@ -370,7 +363,7 @@ impl FilterPlan {
                     }
                 })
             })
-            .collect::<PolarResult<Vec<ResultSet>>>()?;
+            .collect::<Result<Vec<ResultSet>>>()?;
 
         Ok(FilterPlan { result_sets }.optimize(explain))
     }
@@ -470,7 +463,7 @@ impl ResultSet {
         }
     }
 
-    fn build(types: &Types, vars: &Vars, this_type: &str) -> PolarResult<Self> {
+    fn build(types: &Types, vars: &Vars, this_type: &str) -> Result<Self> {
         let result_set = ResultSet {
             requests: HashMap::new(),
             resolve_order: vec![],
@@ -514,7 +507,7 @@ impl FetchRequest {
 }
 
 impl<'a> ResultSetBuilder<'a> {
-    fn into_result_set(self) -> PolarResult<ResultSet> {
+    fn into_result_set(self) -> Result<ResultSet> {
         let mut rset = self.result_set;
         for (i, rid1) in rset.resolve_order.iter().enumerate() {
             let ro = &rset.resolve_order;
@@ -533,24 +526,24 @@ impl<'a> ResultSetBuilder<'a> {
 
         // error messages
         let missing = |id| {
-            err_invalid(format!(
+            invalid_state_error(format!(
                 "Request {} missing from resolve order {:?}",
                 id, order
             ))
         };
         let bad_order = |id1, id2, rset| {
-            err_invalid(format!(
+            invalid_state_error(format!(
                 "Result set {} is resolved before its dependency {} in {:?}",
                 id1, id2, rset
             ))
         };
 
         for (id1, v) in rset.requests.iter() {
-            match index_of(order, id1) {
+            match order.iter().position(|i| i == id1) {
                 None => return missing(*id1),
                 Some(idx1) => {
                     for id2 in v.deps() {
-                        match index_of(order, &id2) {
+                        match order.iter().position(|i| *i == id2) {
                             None => return missing(id2),
                             Some(idx2) if idx2 >= idx1 => return bad_order(id1, id2, &rset),
                             _ => (),
@@ -562,7 +555,7 @@ impl<'a> ResultSetBuilder<'a> {
         Ok(rset)
     }
 
-    fn constrain_var(&mut self, id: Id, var_type: &str) -> PolarResult<&mut Self> {
+    fn constrain_var(&mut self, id: Id, var_type: &str) -> Result<&mut Self> {
         if self.seen.insert(id) {
             // add a fetch request
             self.result_set.requests.insert(
@@ -585,7 +578,7 @@ impl<'a> ResultSetBuilder<'a> {
         Ok(self)
     }
 
-    fn constrain_neq_vars(&mut self, id: Id) -> PolarResult<&mut Self> {
+    fn constrain_neq_vars(&mut self, id: Id) -> Result<&mut Self> {
         let request = self.result_set.requests.get_mut(&id).unwrap();
         for v in self.vars.uncycles.get(&id).into_iter().flatten() {
             let (kind, value) = if let Some(val) = self.vars.eq_values.get(v) {
@@ -604,7 +597,7 @@ impl<'a> ResultSetBuilder<'a> {
         Ok(self)
     }
 
-    fn constrain_eq_vars(&mut self, id: Id) -> PolarResult<&mut Self> {
+    fn constrain_eq_vars(&mut self, id: Id) -> Result<&mut Self> {
         if let Some(t) = self.vars.eq_values.get(&id) {
             self.result_set.requests.get_mut(&id).unwrap().constrain(
                 ConstraintKind::Eq,
@@ -615,7 +608,7 @@ impl<'a> ResultSetBuilder<'a> {
         Ok(self)
     }
 
-    fn constrain_in_vars(&mut self, id: Id, var_type: &str) -> PolarResult<&mut Self> {
+    fn constrain_in_vars(&mut self, id: Id, var_type: &str) -> Result<&mut Self> {
         let mut req = self.result_set.requests.remove(&id).unwrap();
 
         // Constrain any vars that are `in` this var.
@@ -643,7 +636,7 @@ impl<'a> ResultSetBuilder<'a> {
         Ok(self)
     }
 
-    fn constrain_field_eq(&mut self, id: Id, field: &str, child: Id) -> PolarResult<&mut Self> {
+    fn constrain_field_eq(&mut self, id: Id, field: &str, child: Id) -> Result<&mut Self> {
         if let Some(v) = self.vars.eq_values.get(&child) {
             self.result_set.requests.get_mut(&id).unwrap().constrain(
                 ConstraintKind::Eq,
@@ -654,7 +647,7 @@ impl<'a> ResultSetBuilder<'a> {
         Ok(self)
     }
 
-    fn constrain_field_neq(&mut self, id: Id, field: &str, child: Id) -> PolarResult<&mut Self> {
+    fn constrain_field_neq(&mut self, id: Id, field: &str, child: Id) -> Result<&mut Self> {
         let req = self.result_set.requests.get_mut(&id).unwrap();
         for other_id in self.vars.uncycles.get(&child).into_iter().flatten() {
             match (
@@ -684,12 +677,7 @@ impl<'a> ResultSetBuilder<'a> {
         Ok(self)
     }
 
-    fn constrain_field_contained(
-        &mut self,
-        id: Id,
-        field: &str,
-        child: Id,
-    ) -> PolarResult<&mut Self> {
+    fn constrain_field_contained(&mut self, id: Id, field: &str, child: Id) -> Result<&mut Self> {
         let request = self.result_set.requests.get_mut(&id).unwrap();
         self.vars
             .contained_values
@@ -711,7 +699,7 @@ impl<'a> ResultSetBuilder<'a> {
         id: Id,
         my_field: &str,
         my_child: Id,
-    ) -> PolarResult<&mut Self> {
+    ) -> Result<&mut Self> {
         for (other_field, other_child) in self
             .vars
             .field_relationships
@@ -736,7 +724,7 @@ impl<'a> ResultSetBuilder<'a> {
         id: Id,
         my_field: &str,
         my_child: Id,
-    ) -> PolarResult<&mut Self> {
+    ) -> Result<&mut Self> {
         for (other_parent, other_children) in self
             .vars
             .field_relationships
@@ -785,7 +773,7 @@ impl<'a> ResultSetBuilder<'a> {
         other_class_tag: &str,
         my_field: &str,
         other_field: &str,
-    ) -> PolarResult<&mut Self> {
+    ) -> Result<&mut Self> {
         self.constrain_var(child, other_class_tag)?
             .result_set
             .requests
@@ -809,12 +797,12 @@ impl<'a> ResultSetBuilder<'a> {
         field: &str,
         child: Id,
         before: usize,
-    ) -> PolarResult<&mut Self> {
+    ) -> Result<&mut Self> {
         let after = self.result_set.requests.get(&id).unwrap().len();
         if before != after {
             Ok(self)
         } else {
-            err_invalid(format!(
+            invalid_state_error(format!(
                 "Unsupported field access: {}.{} = {}",
                 self.var_name(id)
                     .unwrap_or_else(|| Symbol(format!("{}", id))),
@@ -834,7 +822,7 @@ impl<'a> ResultSetBuilder<'a> {
         })
     }
 
-    fn constrain_fields(&mut self, id: Id, var_type: &str) -> PolarResult<&mut Self> {
+    fn constrain_fields(&mut self, id: Id, var_type: &str) -> Result<&mut Self> {
         match self.vars.field_relationships.get(&id) {
             None => Ok(self),
             Some(fs) => fs.iter().fold(Ok(self), |this, (field, child)| {
@@ -865,13 +853,13 @@ impl<'a> ResultSetBuilder<'a> {
 }
 
 impl Vars {
-    fn from_op(op: &Operation) -> PolarResult<Self> {
+    fn from_op(op: &Operation) -> Result<Self> {
         Self::from_info(VarInfo::from_op(op)?)
     }
 
     /// Collapses the var info that we obtained from walking the expressions.
     /// Track equivalence classes of variables and assign each one an id.
-    fn from_info(info: VarInfo) -> PolarResult<Self> {
+    fn from_info(info: VarInfo) -> Result<Self> {
         let counter = info.counter;
 
         // group the variables into equivalence classes.
@@ -925,7 +913,7 @@ impl Vars {
         });
 
         seek_var_id(&variables, &sym!("_this")).map_or_else(
-            || err_invalid("No `_this` variable".to_string()),
+            || invalid_state_error("No `_this` variable".to_string()),
             |this_id| {
                 Ok(Vars {
                     variables,
@@ -1002,9 +990,7 @@ fn seek_var_id(vars: &HashMap<Id, HashSet<VarName>>, var: &VarName) -> Option<Id
 fn get_var_id(vars: &mut HashMap<Id, HashSet<VarName>>, var: VarName, counter: &Counter) -> Id {
     seek_var_id(vars, &var).unwrap_or_else(|| {
         let new_id = counter.next();
-        let mut new_set = HashSet::new();
-        new_set.insert(var);
-        vars.insert(new_id, new_set);
+        vars.insert(new_id, singleton(var));
         new_id
     })
 }
@@ -1030,13 +1016,6 @@ where
         })
 }
 
-fn index_of<A>(v: &[A], x: &A) -> Option<usize>
-where
-    A: PartialEq<A>,
-{
-    v.iter().position(|y| y == x)
-}
-
 fn hash_map_set_add<A, B>(mut map: HashMap<A, HashSet<B>>, a: A, b: B) -> HashMap<A, HashSet<B>>
 where
     A: Eq + Hash,
@@ -1049,11 +1028,9 @@ where
 #[cfg(test)]
 mod test {
     use super::*;
-    use crate::{
-        bindings::Bindings,
-        error::{ErrorKind::*, OperationalError::Unimplemented, PolarError},
-    };
-    type TestResult = PolarResult<()>;
+    use crate::{bindings::Bindings, error::RuntimeError::*};
+
+    type TestResult = Result<()>;
 
     impl From<Bindings> for ResultEvent {
         fn from(bindings: Bindings) -> Self {
@@ -1074,13 +1051,12 @@ mod test {
         b.is_empty()
     }
 
-    #[test]
-    fn test_dot_plan() -> TestResult {
+    fn test_input_0() -> Term {
         let ins0: Term = ExternalInstance::from(0).into();
         let ins1: Term = ExternalInstance::from(1).into();
         let pat_a = term!(pattern!(instance!("A")));
         let pat_b = term!(pattern!(instance!("B")));
-        let partial = term!(op!(
+        term!(op!(
             And,
             term!(op!(Isa, var!("_this"), pat_a)),
             term!(op!(Isa, ins0.clone(), pat_b.clone())),
@@ -1095,8 +1071,12 @@ mod test {
                 term!(op!(Dot, var!("_this"), str!("field"))),
                 ins1
             ))
-        ));
+        ))
+    }
 
+    #[test]
+    fn test_dot_plan() -> TestResult {
+        let partial = test_input_0();
         let bindings = ResultEvent::from(hashmap! {
             sym!("resource") => partial
         });
@@ -1150,10 +1130,8 @@ mod test {
 
         let err = build_filter_plan(types, vec![bindings], "resource", "A").unwrap_err();
         match err {
-            PolarError {
-                kind: Runtime(RuntimeError::DataFilteringFieldMissing { var_type, field }),
-                ..
-            } if var_type == "A" && field == "field" => (),
+            RuntimeError::DataFilteringFieldMissing { var_type, field }
+                if var_type == "A" && field == "field" => {}
             _ => panic!("unexpected {:?}", err),
         }
         Ok(())
@@ -1167,7 +1145,7 @@ mod test {
     }
 
     #[test]
-    fn test_dot_var_cycles() -> PolarResult<()> {
+    fn test_dot_var_cycles() -> Result<()> {
         let dot_op: Term = term!(op!(Dot, var!("x"), str!("y")));
         let op = op!(
             And,
@@ -1188,13 +1166,13 @@ mod test {
     fn test_unsupported_op_msgs() {
         let err = Vars::from_op(&op!(Dot)).expect_err("should've failed");
         match err {
-            PolarError {
-                kind: Operational(Unimplemented { msg }),
-                ..
-            } => assert_eq!(
-                &msg,
-                "the expression Dot/0 is not supported for data filtering"
-            ),
+            DataFilteringUnsupportedOp {
+                operation:
+                    Operation {
+                        operator: Operator::Dot,
+                        args,
+                    },
+            } if args.is_empty() => (),
             _ => panic!("unexpected"),
         }
     }
