@@ -172,7 +172,6 @@ pub fn to_polar_parens(op: Operator, t: &Term) -> String {
 }
 
 pub mod display {
-    use crate::formatting::{format_args, format_params};
     use std::fmt;
     use std::sync::Arc;
 
@@ -181,7 +180,7 @@ pub mod display {
     use crate::numerics::Numeric;
     use crate::resource_block::Declaration;
     use crate::rules::Rule;
-    use crate::terms::{Operation, Operator, Symbol, Term, Value};
+    use crate::terms::{Call, Operator, Symbol, Term};
     use crate::vm::*;
 
     impl fmt::Display for Binding {
@@ -202,6 +201,18 @@ pub mod display {
         }
     }
 
+    impl fmt::Display for Call {
+        fn fmt(&self, fmt: &mut fmt::Formatter<'_>) -> fmt::Result {
+            write!(fmt, "{}", self.to_polar())
+        }
+    }
+
+    impl fmt::Display for Operator {
+        fn fmt(&self, fmt: &mut fmt::Formatter<'_>) -> fmt::Result {
+            write!(fmt, "{}", self.to_polar())
+        }
+    }
+
     impl fmt::Display for Choice {
         fn fmt(&self, fmt: &mut fmt::Formatter<'_>) -> fmt::Result {
             write!(
@@ -314,30 +325,7 @@ pub mod display {
 
     impl fmt::Display for Rule {
         fn fmt(&self, fmt: &mut fmt::Formatter) -> fmt::Result {
-            match &self.body.value() {
-                Value::Expression(Operation {
-                    operator: Operator::And,
-                    args,
-                }) => {
-                    if args.is_empty() {
-                        write!(
-                            fmt,
-                            "{}({});",
-                            self.name.to_polar(),
-                            format_params(&self.params, ", ")
-                        )
-                    } else {
-                        write!(
-                            fmt,
-                            "{}({}) if {};",
-                            self.name.to_polar(),
-                            format_params(&self.params, ", "),
-                            format_args(Operator::And, args, ",\n  "),
-                        )
-                    }
-                }
-                _ => panic!("Not any sorta rule I parsed"),
-            }
+            write!(fmt, "{}", self.to_polar())
         }
     }
 
@@ -349,6 +337,7 @@ pub mod display {
             }
         }
     }
+
     impl fmt::Display for Declaration {
         fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
             match self {
@@ -358,6 +347,16 @@ pub mod display {
             }
         }
     }
+
+    impl fmt::Display for LogLevel {
+        fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+            match self {
+                Self::Trace => write!(f, "trace"),
+                Self::Debug => write!(f, "debug"),
+                Self::Info => write!(f, "info"),
+            }
+        }
+    }
 }
 
 pub mod to_polar {
@@ -387,12 +386,20 @@ pub mod to_polar {
     impl ToPolarString for ExternalInstance {
         fn to_polar(&self) -> String {
             if let Some(ref repr) = self.repr {
-                repr.clone()
+                format!(
+                    "{} TYPE `{}`",
+                    repr.clone(),
+                    self.class_repr.as_ref().unwrap_or(&"UNKNOWN".to_string())
+                )
             } else {
                 // Print out external instances like ^{id: 123}
                 // NOTE: this format is used by host libraries to enrich output
                 // messages with native representations of the instances.
-                format!("^{{id: {}}}", self.instance_id)
+                format!(
+                    "^{{id: {}}} TYPE `{}`",
+                    self.instance_id,
+                    self.class_repr.as_ref().unwrap_or(&"UNKNOWN".to_string())
+                )
             }
         }
     }
@@ -560,6 +567,12 @@ pub mod to_polar {
         }
     }
 
+    impl Rule {
+        pub fn head_as_string(&self) -> String {
+            format!("{}({})", self.name, format_params(&self.params, ", "))
+        }
+    }
+
     impl ToPolarString for Rule {
         fn to_polar(&self) -> String {
             match &self.body.value() {
@@ -568,16 +581,11 @@ pub mod to_polar {
                     args,
                 }) => {
                     if args.is_empty() {
-                        format!(
-                            "{}({});",
-                            self.name.to_polar(),
-                            format_params(&self.params, ", ")
-                        )
+                        format!("{};", self.head_as_string())
                     } else {
                         format!(
-                            "{}({}) if {};",
-                            self.name.to_polar(),
-                            format_params(&self.params, ", "),
+                            "{} if {};",
+                            self.head_as_string(),
                             format_args(Operator::And, args, " and "),
                         )
                     }
