@@ -12,6 +12,7 @@ pub const FALSE: Operation = op!(Or);
 
 /// Invert operators.
 pub fn invert_operation(Operation { operator, args }: Operation) -> Operation {
+    use Operator::*;
     fn invert_args(args: Vec<Term>) -> Vec<Term> {
         args.into_iter()
             .map(|t| {
@@ -23,50 +24,52 @@ pub fn invert_operation(Operation { operator, args }: Operation) -> Operation {
     }
 
     match operator {
-        Operator::And => Operation {
-            operator: Operator::Or,
+        // noop
+        Debug | Print | New | Dot => Operation { operator, args },
+
+        // de morgan
+        And => Operation {
+            operator: Or,
             args: invert_args(args),
         },
-        Operator::Or => Operation {
-            operator: Operator::And,
+        Or => Operation {
+            operator: And,
             args: invert_args(args),
         },
-        Operator::Unify | Operator::Eq => Operation {
-            operator: Operator::Neq,
+
+        // opposite operator
+        Unify | Eq => Operation {
+            operator: Neq,
             args,
         },
-        Operator::Neq => Operation {
-            operator: Operator::Unify,
+        Neq => Operation {
+            operator: Unify,
             args,
         },
-        Operator::Gt => Operation {
-            operator: Operator::Leq,
+        Gt => Operation {
+            operator: Leq,
             args,
         },
-        Operator::Geq => Operation {
-            operator: Operator::Lt,
+        Geq => Operation { operator: Lt, args },
+        Lt => Operation {
+            operator: Geq,
             args,
         },
-        Operator::Lt => Operation {
-            operator: Operator::Geq,
-            args,
-        },
-        Operator::Leq => Operation {
-            operator: Operator::Gt,
-            args,
-        },
-        Operator::Debug | Operator::Print | Operator::New | Operator::Dot => {
-            Operation { operator, args }
-        }
-        Operator::Isa => Operation {
-            operator: Operator::Not,
-            args: vec![term!(op!(Isa, args[0].clone(), args[1].clone()))],
-        },
-        Operator::Not => args[0]
+        Leq => Operation { operator: Gt, args },
+
+        // double negative
+        Not => args[0]
             .value()
             .as_expression()
             .expect("negated expression")
             .clone(),
+
+        // preserve the not
+        Isa | In => Operation {
+            operator: Not,
+            args: vec![term!(Operation { operator, args })],
+        },
+
         _ => todo!("negate {:?}", operator),
     }
 }
@@ -490,7 +493,7 @@ mod test {
         // NOTE(gj): only one permutation remains parse-able.
         p.load_str("m(x) if [_y] matches [x];")?;
         let mut q = p.new_query_from_term(term!(call!("m", [sym!("x")])), false);
-        assert_partial_expression!(next_binding(&mut q)?, "x", "__y_34 matches _this");
+        assert_partial_expression!(next_binding(&mut q)?, "x", "__y_36 matches _this");
         assert_query_done!(q);
 
         // TODO(gj): Make the below work.
@@ -1107,7 +1110,7 @@ mod test {
         assert_partial_expression!(
             next_binding(&mut q),
             "x",
-            "_this matches A{} and _b_74 in _this.b and _b_74 matches B{} and 1 = _b_74.foo"
+            "_this matches A{} and _b_81 in _this.b and _b_81 matches B{} and 1 = _b_81.foo"
         );
         assert_query_done!(q);
 
@@ -1115,7 +1118,7 @@ mod test {
         assert_partial_expression!(
             next_binding(&mut q),
             "x",
-            "_this matches A{} and _b_84 in _this.b and _b_84 matches B{} and _b_84.c matches C{} and 1 = _b_84.c.bar"
+            "_this matches A{} and _b_92 in _this.b and _b_92 matches B{} and _b_92.c matches C{} and 1 = _b_92.c.bar"
         );
         assert_query_done!(q);
 
@@ -1123,7 +1126,7 @@ mod test {
         assert_partial_expression!(
             next_binding(&mut q),
             "x",
-            "_this matches A{} and _b_104 in _this.b and _b_104.c matches C{} and 1 = _b_104.c.bar"
+            "_this matches A{} and _b_115 in _this.b and _b_115.c matches C{} and 1 = _b_115.c.bar"
         );
         // @TODO(sam): this result is incorrect. We *could* know
         // that `_b_104` matches B{} by checking `a.b` first
@@ -1131,7 +1134,7 @@ mod test {
         assert_partial_expression!(
             next_binding(&mut q),
             "x",
-            "_this matches A{} and _b_104 in _this.b and _b_104.c matches D{} and 2 = _b_104.c.bar"
+            "_this matches A{} and _b_115 in _this.b and _b_115.c matches D{} and 2 = _b_115.c.bar"
         );
         assert_query_done!(q);
         Ok(())
@@ -1215,7 +1218,7 @@ mod test {
         p.register_constant(sym!("y"), term!(value!(op!(And))))?;
         let mut q = p.new_query_from_term(term!(call!("f", [sym!("x"), sym!("y")])), false);
         let next = next_binding(&mut q)?;
-        assert_partial_expressions!(next, "x" => "_this = y", "y" => "x = _this");
+        assert_eq!(next[&sym!("x")], var!("y"));
         let next = next_binding(&mut q)?;
         assert_eq!(next[&sym!("x")], term!(1));
         assert_eq!(next[&sym!("y")], term!(1));
