@@ -821,6 +821,7 @@ mod tests {
             instance_id: 1,
             constructor: None,
             repr: None,
+            class_repr: None,
         };
         let repo_term = term!(Value::ExternalInstance(repo_instance.clone()));
         let repo_name = sym!("Repo");
@@ -976,6 +977,7 @@ mod tests {
             instance_id: 1,
             constructor: None,
             repr: None,
+            class_repr: None,
         };
         let repo_term = term!(Value::ExternalInstance(repo_instance.clone()));
         let repo_name = sym!("Repo");
@@ -1061,6 +1063,7 @@ mod tests {
             instance_id: 2,
             constructor: None,
             repr: None,
+            class_repr: None,
         };
         let user_term = term!(Value::ExternalInstance(user_instance.clone()));
         let user_name = sym!("User");
@@ -1531,7 +1534,8 @@ mod tests {
             term!(Value::ExternalInstance(ExternalInstance {
                 instance_id: 1,
                 constructor: None,
-                repr: None
+                repr: None,
+                class_repr: None,
             })),
         )
         .unwrap();
@@ -1540,7 +1544,8 @@ mod tests {
             term!(Value::ExternalInstance(ExternalInstance {
                 instance_id: 2,
                 constructor: None,
-                repr: None
+                repr: None,
+                class_repr: None,
             })),
         )
         .unwrap();
@@ -1549,7 +1554,8 @@ mod tests {
             term!(Value::ExternalInstance(ExternalInstance {
                 instance_id: 3,
                 constructor: None,
-                repr: None
+                repr: None,
+                class_repr: None,
             })),
         )
         .unwrap();
@@ -1564,7 +1570,8 @@ mod tests {
             term!(Value::ExternalInstance(ExternalInstance {
                 instance_id: 4,
                 constructor: None,
-                repr: None
+                repr: None,
+                class_repr: None,
             })),
         )
         .unwrap();
@@ -1710,6 +1717,7 @@ mod tests {
             instance_id: 1,
             constructor: None,
             repr: None,
+            class_repr: None,
         };
         let repo_term = term!(Value::ExternalInstance(repo_instance.clone()));
         let repo_name = sym!("Repository");
@@ -1720,6 +1728,7 @@ mod tests {
             instance_id: 2,
             constructor: None,
             repr: None,
+            class_repr: None,
         };
         let org_term = term!(Value::ExternalInstance(org_instance.clone()));
         let org_name = sym!("Organization");
@@ -1764,6 +1773,7 @@ mod tests {
             instance_id: 1,
             constructor: None,
             repr: None,
+            class_repr: None,
         };
         let team_term = term!(Value::ExternalInstance(team_instance.clone()));
         let team_name = sym!("Team");
