@@ -51,6 +51,14 @@ impl Rule {
         }
     }
 
+    pub fn get_source_id(&self) -> Option<u64> {
+        if let SourceInfo::Parser { src_id, .. } = self.source_info {
+            Some(src_id)
+        } else {
+            None
+        }
+    }
+
     pub fn new_from_test(name: Symbol, params: Vec<Parameter>, body: Term) -> Self {
         Self {
             name,
@@ -99,8 +107,6 @@ impl RuleTypes {
     fn add_default_rule_types(&mut self) {
         // type has_permission(actor: Actor, permission: String, resource: Resource);
         self.add(rule!("has_permission", ["actor"; instance!(sym!("Actor")), "_permission"; instance!(sym!("String")), "resource"; instance!(sym!("Resource"))]));
-        // type has_permission(actor: Actor, permission: String, resource: Actor);
-        self.add(rule!("has_permission", ["actor"; instance!(sym!("Actor")), "_permission"; instance!(sym!("String")), "resource"; instance!(sym!("Actor"))]));
         // type allow(actor, action, resource);
         self.add(rule!(
             "allow",
