@@ -414,6 +414,37 @@ fn test_jealous() -> TestResult {
     Ok(())
 }
 
+#[test]
+fn test_forall_with_dot_lookup() -> TestResult {
+    let p = polar();
+    p.load_str(
+        r#"
+        foo(x) if forall(y in x.ys, y != 0);
+        moo({ys}) if forall(y in ys, y != 0);
+        goo(x) if forall(x matches {ys} and y in ys, y != 0);
+        boo(x) if forall(y in x.ys, y != 0 and y != 1);
+    "#,
+    )?;
+
+    qeval(&p, "foo({ys: []})");
+    qeval(&p, "moo({ys: []})");
+    qeval(&p, "goo({ys: []})");
+
+    qeval(&p, "foo({ys: [9, 8, 7]})");
+    qeval(&p, "moo({ys: [8, 7, 6]})");
+    qeval(&p, "goo({ys: [7, 8, 9]})");
+
+    qnull(&p, "foo({ys: [-1, 0, 1]})");
+    qnull(&p, "moo({ys: [0, 1, 2]})");
+    qnull(&p, "goo({ys: [2, 1, 0]})");
+
+    qeval(&p, "boo({ys: []})");
+    qeval(&p, "boo({ys: [9, 8, 7]})");
+    qnull(&p, "boo({ys: [-2, -1, 0]})");
+
+    Ok(())
+}
+
 #[test]
 fn test_trace() -> TestResult {
     let p = polar();
@@ -1612,6 +1643,7 @@ fn test_missing_resource_hint() -> TestResult {
         instance_id: 1,
         constructor: None,
         repr: None,
+        class_repr: None,
     };
     let repo_term = term!(Value::ExternalInstance(repo_instance.clone()));
     let repo_name = sym!("Repository");
@@ -1622,6 +1654,7 @@ fn test_missing_resource_hint() -> TestResult {
         instance_id: 2,
         constructor: None,
         repr: None,
+        class_repr: None,
     };
     let organization_term = term!(Value::ExternalInstance(organization_instance.clone()));
     let organization_name = sym!("Organization");
@@ -1632,6 +1665,7 @@ fn test_missing_resource_hint() -> TestResult {
         instance_id: 3,
         constructor: None,
         repr: None,
+        class_repr: None,
     };
     let user_term = term!(Value::ExternalInstance(user_instance.clone()));
     let user_name = sym!("User");
@@ -2097,6 +2131,16 @@ fn test_cut() -> TestResult {
     Ok(())
 }
 
+#[test]
+fn test_forall_with_dots_on_rhs() -> TestResult {
+    let p = polar();
+    p.load_str("goo(x, y) if forall(z in y, z.n < x);")?;
+    qeval(&p, "goo(10, [])");
+    qeval(&p, "goo(10, [{n:1}, {n:2}, {n:3}])");
+    qnull(&p, "goo(10, [{n:11}, {n:10}, {n:9}])");
+    Ok(())
+}
+
 #[test]
 fn test_forall() -> TestResult {
     let p = polar();
@@ -2137,6 +2181,7 @@ fn test_forall() -> TestResult {
            test(_: {z: 1}, y) if y = 3;"#,
     )?;
     qeval(&p, "forall(test({x: 1, y: 1, z: 1}, y), y in [1, 2, 3])");
+
     Ok(())
 }
 
@@ -2518,6 +2563,7 @@ fn test_suggested_rule_specializer() -> TestResult {
         instance_id: 1,
         constructor: None,
         repr: None,
+        class_repr: None,
     };
     let repo_term = term!(Value::ExternalInstance(repo_instance.clone()));
     let repo_name = sym!("Repository");
@@ -2528,6 +2574,7 @@ fn test_suggested_rule_specializer() -> TestResult {
         instance_id: 2,
         constructor: None,
         repr: None,
+        class_repr: None,
     };
     let user_term = term!(Value::ExternalInstance(user_instance.clone()));
     let user_name = sym!("User");
@@ -2569,6 +2616,7 @@ fn test_missing_required_rule_type() -> TestResult {
         instance_id: 1,
         constructor: None,
         repr: None,
+        class_repr: None,
     };
     let repo_term = term!(Value::ExternalInstance(repo_instance.clone()));
     let repo_name = sym!("Repository");
@@ -2579,6 +2627,7 @@ fn test_missing_required_rule_type() -> TestResult {
         instance_id: 2,
         constructor: None,
         repr: None,
+        class_repr: None,
     };
     let issue_term = term!(Value::ExternalInstance(issue_instance.clone()));
     let issue_name = sym!("Issue");
@@ -2589,6 +2638,7 @@ fn test_missing_required_rule_type() -> TestResult {
         instance_id: 3,
         constructor: None,
         repr: None,
+        class_repr: None,
     };
     let user_term = term!(Value::ExternalInstance(user_instance.clone()));
     let user_name = sym!("User");
